import socket
import ssl
import sys
import task_information
import time
import json
import base64
import threading
import bcrypt
import hashlib
import secrets
import client_crypto

memory_lock = threading.RLock()

_version = "1.7.1"
def get_version():
    return _version

# --- not recommended to use outside this file --- #
default_server_socket: ssl.SSLSocket = None

# this function will initialize a server_socket.
# After calling this function you don't need to tell the other functions the server_socket anymore. But you still can.
def initialize_server_socket(server_socket: ssl.SSLSocket):
    global default_server_socket
    default_server_socket = server_socket

# --- not recommended to get called outside this file --- #
def recv_msg_in_bytes(server_socket: ssl.SSLSocket = None) -> bytes:
    if server_socket == None: server_socket = default_server_socket

    data = server_socket.recv(1024)
    number_of_bytes = int(str(data, 'utf8'))
    if number_of_bytes == 0:
        number_of_bytes = 16
    data = server_socket.recv(number_of_bytes)
    number_of_bytes = int(str(data, 'utf8'))

    if number_of_bytes > 4096:
        last_chunk_size = number_of_bytes % 4096
        number_of_chunks = number_of_bytes // 4096
        msg = b''
        for _ in range(number_of_chunks):
            msg += server_socket.recv(4096)
        if last_chunk_size != 0:
            msg += server_socket.recv(last_chunk_size)
    else:
        msg = server_socket.recv(number_of_bytes)
    
    return msg

# --- not recommended to get called outside this file --- #
def recv_msg(server_socket: ssl.SSLSocket = None) -> str:
    if server_socket == None: server_socket = default_server_socket

    return str(recv_msg_in_bytes(server_socket=server_socket), 'utf8')

# --- not recommended to get called outside this file --- #
def recv_msg_as_list(server_socket: ssl.SSLSocket = None) -> list:
    if server_socket == None: server_socket = default_server_socket

    return list(recv_msg_in_bytes(server_socket=server_socket))

# --- not recommended to get called outside this file --- #
def recv_msg_as_json(server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    return json.loads(recv_msg(server_socket=server_socket))

# --- not recommended to get called outside this file --- #
def send_msg_in_bytes(msg: bytes, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket
    
    length = len(msg)
    length_bytes = bytes(str(length), 'utf8')
    if len(length_bytes) == 1:
        server_socket.send(length_bytes)
    else:
        server_socket.send(b'0')
        length_to_send = b'0' * (16 - len(length_bytes)) + length_bytes
        server_socket.send(length_to_send)
    if length > 4096:
        number_of_chunks = length // 4096
        last_chunk_size = length % 4096
        for i in range(number_of_chunks):
            val = i * 4096
            bytes_to_send = msg[val:val+4096]
            server_socket.send(bytes_to_send)
        if last_chunk_size != 0:
            server_socket.send(msg[-last_chunk_size:])
    else:
        server_socket.send(msg)


# --- not recommended to get called outside this file --- #
def send_msg(msg: str, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    b_msg = bytes(msg, 'utf8')
    send_msg_in_bytes(server_socket=server_socket, msg=b_msg)

# --- not recommended to get called outside this file --- #
def send_msg_as_list(msg: list, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    l_msg = bytes(msg)
    send_msg_in_bytes(server_socket=server_socket, msg=l_msg)

# --- not recommended to get called outside this file --- #
def send_msg_as_json(msg: dict, indent = None, sort_keys=False, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    json_msg = json.dumps(msg, indent=indent, sort_keys=sort_keys)
    send_msg(msg=json_msg, server_socket=server_socket)

# --- not recommended to get called outside this file --- #
def send_task(task: list, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    json_dict = {"t": task[0]}
    if len(task) > 1:
        json_dict.update({"p": task[1:]})

    send_msg_as_json(server_socket=server_socket, msg=json_dict)

# the server will not return "result" here. Instead of that he will return "valid" as key and a boolean as value
def start_task_api_version_check(server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    json_dict = {"v": get_version()}

    send_msg_as_json(msg=json_dict, server_socket=server_socket)
    ret_val = recv_msg_as_json(server_socket=server_socket)
    return ret_val

# this function will return an integer, which tells you the outcome of the task:
# 0    successfull. You are now logged in.
# 1    Something went wrong. Probably the username or password is wrong.
# 2    Already logged in as another user.
def start_task_login(username: str, password: str, task_parameters: list = task_information.login.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.login.task_number] + task_parameters
    send_task(task=task_list, server_socket=server_socket)

    # hashing password
    pwd_bytes = bytes(password, 'utf8')
    sha256_pwd = base64.b64encode(hashlib.sha256(pwd_bytes).digest())
    bcrypt_pwd = bcrypt.hashpw(sha256_pwd, bcrypt.gensalt())
    b64pwd: str = str(base64.b64encode(bcrypt_pwd), 'utf8')

    information_dict = {"u": username, "p": b64pwd}
    send_msg_as_json(msg=information_dict, server_socket=server_socket)

    # return the outcome of the task
    task_result: dict = recv_msg_as_json(server_socket=server_socket)
    return task_result
    
# this function will return an integer, which tells you the outcome of the task:
# 0    successfull. You are now logged in.
# 1    Something went wrong.
# 2    Something went wrong. The username is probably already given or not usable for some reason.
# 3    Password is too short. You need a minimum of 8 characters
def start_task_register(username: str, password: str,
 registration_id: int = 0,  # placeholder
 device_id: int = 0,  # placeholder
 identity_key_public: bytes = b"",  # placeholder
 identity_key_private: bytes = b"",  # placeholder
 signed_pre_key_public: bytes = b"",  # placeholder
 signed_pre_key_private: bytes = b"",  # placeholder
 signed_pre_key_signature: bytes = b"",  # placeholder
 pre_keys_public: list = [],  # placeholder
 pre_keys_private: list = [],  # placeholder
 task_parameters: list = task_information.register.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.register.task_number] + task_parameters
    send_task(task=task_list, server_socket=server_socket)

    # hashing password
    pwd_bytes = bytes(password, 'utf8')
    sha256_pwd = base64.b64encode(hashlib.sha256(pwd_bytes).digest())
    bcrypt_pwd = bcrypt.hashpw(sha256_pwd, bcrypt.gensalt())
    b64pwd: str = str(base64.b64encode(bcrypt_pwd), 'utf8')

    # Creation of the standard AES key of the client
    key: bytes = secrets.token_bytes(32) # creating key of 32 bytes (256 Bit) with a cryptographically secure pseudorandom function
    encrypted_key, encrypted_key_salt = client_crypto.AES.encrypt(key, pwd_bytes)
    b64encrypted_key: str = str(base64.b64encode(encrypted_key), 'utf8')
    b64encrypted_key_salt: str = str(base64.b64encode(encrypted_key_salt), 'utf8')

    # signal identity key pair
    b64_identity_key_public: str = str(base64.b64encode(identity_key_public), 'utf8')
    encrypted_identity_key_private, encrypted_identity_key_private_salt = client_crypto.AES.encrypt(identity_key_private, key)
    b64_encrypted_identity_key_private: str = str(base64.b64encode(encrypted_identity_key_private), 'utf8')
    b64_encrypted_identity_key_private_salt: str = str(base64.b64encode(encrypted_identity_key_private_salt), 'utf8')

    # signal signed pre key pair
    b64_signed_pre_key_public: str = str(base64.b64encode(signed_pre_key_public), 'utf8')
    b64_signed_pre_key_signature: str = str(base64.b64encode(signed_pre_key_signature), 'utf8')
    encrypted_signed_pre_key_private, encrypted_signed_pre_key_private_salt = client_crypto.AES.encrypt(signed_pre_key_private, key)
    b64_encrypted_signed_pre_key_private: str = str(base64.b64encode(encrypted_signed_pre_key_private), 'utf8')
    b64_encrypted_signed_pre_key_private_salt: str = str(base64.b64encode(encrypted_signed_pre_key_private_salt), 'utf8')

    # signal pre key pairs
    b64_pre_keys_public: list = []
    b64_encrypted_pre_keys_private: list = []
    b64_encrypted_pre_keys_salts = []
    for i in pre_keys_public:
        b64_pre_keys_public.append(str(base64.b64encode(i), 'utf8'))
    for i in pre_keys_private:
        encrypted_key, salt = client_crypto.AES.encrypt(i, key)
        b64_encrypted_pre_keys_private.append(str(base64.b64encode(encrypted_key), 'utf8'))
        b64_encrypted_pre_keys_salts.append(str(base64.b64encode(salt), 'utf8'))

    information_dict = {"u": username,
                        "p": b64pwd,
                        "k": b64encrypted_key,
                        "ki": b64encrypted_key_salt,
                        "r": registration_id,
                        "d": device_id,
                        "i1": b64_identity_key_public,
                        "i2": b64_encrypted_identity_key_private,
                        "ii": b64_encrypted_identity_key_private_salt,
                        "s1": b64_signed_pre_key_public,
                        "s2": b64_encrypted_signed_pre_key_private,
                        "s3": b64_signed_pre_key_signature,
                        "si": b64_encrypted_signed_pre_key_private_salt,
                        "p1": b64_pre_keys_public,
                        "p2": b64_encrypted_pre_keys_private,
                        "pi": b64_encrypted_pre_keys_salts}
    send_msg_as_json(msg=information_dict, server_socket=server_socket)

    # return the outcome of the task
    task_result: dict = recv_msg_as_json(server_socket=server_socket)
    return task_result

# this function will return an integer, which tells you the outcome of the task:
# 0    successfull. You are no longer logged in.
def start_task_logout(task_parameters: list = task_information.logout.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.logout.task_number] + task_parameters
    send_task(task=task_list, server_socket=server_socket)

    # return the outcome of the task
    task_result: dict = recv_msg_as_json(server_socket=server_socket)
    return task_result

# this function will return a string. The first index of the string will give you information about the outcome of the task, formatted as a binary number. So you can't just convert to int:
# 0     Everything was successfull
# 1     Undefined Error
# 2     Client isn't logged in
# 3     Label invalid
# The rest of the string can be some additional information
# Please be carefull to use the parameters correctly. Its all about repeat.
# If repeat is true, you need to give information for the start_point, stop_point and frequence, but not for the time_point. The start_point and stop_point can be left out. If so, it will start immediately and never stop. The frequence has to be given.
# If repeat is false, you don't need to give information for the start_point, stop_point and frequence, but for the time_point.
def start_task_memory(key: bytes = None,
                    label: str = None, msg: str = None,
                    receiver: str = "self", repeat: bool = False,
                    start_point: float = None, stop_point: float = None,
                    frequence: float = None, time_point = None,
                    information: str = None,
                    task_parameters: list = task_information.memory.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    global memory_lock
    if msg is not None:
        encrypted_msg, encrypted_msg_salt = client_crypto.AES.encrypt(bytes(msg, 'utf8'), key)
        b64_encrypted_msg: str = str(base64.b64encode(encrypted_msg), 'utf8')
        b64_encrypted_msg_salt: str = str(base64.b64encode(encrypted_msg_salt), 'utf8')
        msg = b64_encrypted_msg
        msg_salt = b64_encrypted_msg_salt
    if label is not None:
        encrypted_label, encrypted_label_salt = client_crypto.AES.encrypt(bytes(label, 'utf8'), key)
        b64_encrypted_label: str = str(base64.b64encode(encrypted_label), 'utf8')
        b64_encrypted_label_salt: str = str(base64.b64encode(encrypted_label_salt), 'utf8')
        label = b64_encrypted_label
        label_salt = b64_encrypted_label_salt
    with memory_lock:
        if server_socket == None: server_socket = default_server_socket

        # starting the task
        task_list: list = [task_information.memory.task_number] + task_parameters
        send_task(server_socket=server_socket, task=task_list)

        information_dict = {}
        # evaluate the task_parameters
        if task_parameters[0] == 0:
            if task_parameters[1] == 0:
                pass # nothing to do there
            elif task_parameters[1] == 1:
                if task_parameters[2] == 0:
                    information_dict.update({"l": label})
                    information_dict.update({"ls": label_salt})
                    send_msg_as_json(msg=information_dict, server_socket=server_socket)
                elif task_parameters[2] == 1:
                    information_dict.update({"l": label, "i": information})
                    information_dict.update({"ls": label_salt})
                    send_msg_as_json(msg=information_dict, server_socket=server_socket)
        elif task_parameters[0] == 1: # We want to send a new memory task to the server
            # sending all information in the right order
            information_dict.update({"l": label})
            information_dict.update({"ls": label_salt})
            information_dict.update({"m": msg})
            information_dict.update({"ms": msg_salt})
            information_dict.update({"r2": receiver})
            information_dict.update({"r1": repeat})

            # We have to send different information based on repeat
            if repeat:
                if start_point == None: start_point = time.time() # giving start_point a value if it doesn't already have one
                if stop_point == None: stop_point: float = 0 # giving stop_point a value if it doesn't already have one
                information_dict.update({"s1": start_point})
                information_dict.update({"s2": stop_point})
                information_dict.update({"f": frequence})
            else:
                information_dict.update({"t": time_point})
            send_msg_as_json(msg=information_dict, server_socket=server_socket)
        elif task_parameters[0] == 2:
            information_dict.update({"l": label})
            information_dict.update({"ls": label_salt})
            send_msg_as_json(msg=information_dict, server_socket=server_socket)
        elif task_parameters[0] == 3:
            information_dict.update({"l": label})
            information_dict.update({"ls": label_salt})
            send_msg_as_json(msg=information_dict, server_socket=server_socket)
        
        # return the outcome of the task
        task_result: dict = recv_msg_as_json(server_socket=server_socket)
        return task_result

# the will return the outcome of the operation in json format (outcome is "result")
# 0     Everything was successfull
# 1     Unexpected Error (The token is probably still there)
# 2     Token invalid
# 3     Wrong format. Error with JSON.
# 4     User not logged in/already logged in
def start_task_login_token(username: str = None, token: str = None, old_token: str = None, task_parameters: list = task_information.login_token.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.login_token.task_number] + task_parameters
    send_task(server_socket=server_socket, task=task_list)

    if task_parameters == task_information.login_token.login_task_parameters:
        information_dict: dict = {"u": username, "t": token}
        send_msg_as_json(information_dict, server_socket=server_socket)
    elif task_parameters == task_information.login_token.get_new_token_task_parameters:
        # returns "result", "token", "start_point" and "stop_point"
        information_dict: dict = {"o": old_token}
        send_msg_as_json(information_dict, server_socket=server_socket)
    elif task_parameters == task_information.login_token.delete_token_task_parameters:
        information_dict: dict = {"t": token}
        send_msg_as_json(information_dict, server_socket=server_socket)
    elif task_parameters == task_information.login_token.delete_all_tokens_task_parameters:
        pass # nothing has to be done here

    # return the outcome of the task
    task_result: dict = recv_msg_as_json(server_socket=server_socket)
    return task_result

def start_task_rsync(device: str, mountpoint: str, task_parameters: list = task_information.rsync.default_task_parameters, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.rsync.task_number] + task_parameters
    send_task(server_socket=server_socket, task=task_list)

    information_dict = {"device": device, "mountpoint": mountpoint}
    send_msg_as_json(information_dict, server_socket=server_socket)

    result_json: dict = recv_msg_as_json(server_socket=server_socket)
    return result_json